from itertools import product, chain
from math import sqrt


class MBR(object):
    """Hold an area in 2D cartesian space.
    """

    def __init__(self, x_min, x_max, y_min, y_max):
        if x_max < x_min:
            raise ValueError('x_max ({}) is less than x_min ({})'.format(x_max, x_min))
        if y_max < y_min:
            raise ValueError('y_max ({}) is less than y_min ({})'.format(y_max, y_min))

        self.x_min = x_min
        self.x_max = x_max
        self.y_min = y_min
        self.y_max = y_max

    @staticmethod
    def create_mbr(x_values, y_values):
        return MBR(min(x_values), max(x_values), min(y_values), max(y_values))

    @staticmethod
    def create_circumventing_mbr(mbrs):
        x_min = min(chain(*((mbr.x_min, mbr.x_max) for mbr in mbrs)))
        x_max = max(chain(*((mbr.x_min, mbr.x_max) for mbr in mbrs)))

        y_min = min(chain(*((mbr.y_min, mbr.y_max) for mbr in mbrs)))
        y_max = max(chain(*((mbr.y_min, mbr.y_max) for mbr in mbrs)))

        return MBR(x_min, x_max, y_min, y_max)

    def corners(self):
        return set(product([self.x_min, self.x_max], [self.y_min, self.y_max]))

    def perimeter(self):
        return 2 * abs(self.x_min-self.x_max) + 2 * abs(self.y_min-self.y_max)

    def area(self):
        """Retuns the size the area defined by the MBR
        """
        return (self.x_max - self.x_min) * (self.y_max - self.y_min)

    def contains_point(self, point):
        """ Checks if a point is contained within the MBR.
        Points that are on border also counts
        """
        return self.x_min <= point[0] <= self.x_max and self.y_min <= point[1] <= self.y_max

    def area_to_circumvent_point(self, point):
        """ Returns the least amount of expanded area to circumvent a
        point.
        """
        new_x_min, _, new_x_max = sorted([self.x_min, self.x_max, point[0]])
        new_y_min, _, new_y_max = sorted([self.y_min, self.y_max, point[1]])
        new_mbr = MBR(new_x_min, new_x_max, new_y_min, new_y_max)
        return new_mbr.area() - self.area()

    def area_to_circumvent_mbr(self, mbr):
        new_x_min, *_, new_x_max = sorted([self.x_min, self.x_max, mbr.x_min, mbr.x_max])
        new_y_min, *_, new_y_max = sorted([self.y_min, self.y_max, mbr.y_min, mbr.y_max])
        new_mbr = MBR(new_x_min, new_x_max, new_y_min, new_y_max)
        return new_mbr.area() - self.area()

    def overlaps(self, mbr):
        overlap_x = False
        overlap_y = False
        if self.x_min <= mbr.x_min <= self.x_max or mbr.x_min <= self.x_min <= mbr.x_max:
            overlap_x = True
        if self.y_min <= mbr.y_min <= self.y_max or mbr.y_min <= self.y_min <= mbr.y_max:
            overlap_y = True
        return overlap_x and overlap_y

    def distance_to(self, mbr):
        if self.overlaps(mbr):
            return 0

        y_overlaps = self.y_min <= mbr.y_min <= self.y_max or mbr.y_min <= self.y_min <= mbr.y_max
        if y_overlaps:
            return min([abs(self.x_min - mbr.x_max), abs(self.x_max - mbr.x_min)])

        x_overlaps = self.x_min <= mbr.x_min <= self.x_max or mbr.x_min <= self.x_min <= mbr.x_max
        if x_overlaps:
            return min([abs(self.y_min - mbr.y_max), abs(self.y_max - mbr.y_min)])

        pairs_of_corners = product(self.corners(), mbr.corners())

        # return euclidian distance
        return min((sqrt((a[0] - b[0])**2 + (a[1] - b[1])**2) for a, b in pairs_of_corners))

    def overlap(self, mbr):
        """
        Returns size of overlap.
        :param mbr:
        :return:
        """
        # depends on overlaps
        if self.overlaps(mbr) is False:
            return 0

        x_values = sorted([self.x_min, self.x_max, mbr.x_min, mbr.x_max])
        y_values = sorted([self.y_min, self.y_max, mbr.y_min, mbr.y_max])

        return MBR(x_values[1], x_values[2], y_values[1], y_values[2]).area()

    def center(self):
        """Returns cordinates (x,y) for center of the mbr"""
        x_center = self.x_min + abs(self.x_min - self.x_max)/2
        y_center = self.y_min + abs(self.y_min - self.y_max)/2

        return x_center, y_center

    def __str__(self):
        return 'x min/max {}/{}, y min/max {}/{}'.format(
            self.x_min,
            self.x_max,
            self.y_min,
            self.y_max
        )

    def __repr__(self):
        return '<MBR({},{},{},{})>'.format(self.x_min, self.x_max, self.y_min, self.y_max)
