from collections import deque
from itertools import combinations, product, chain

from rtree.settings import Settings
from .mbr import MBR
from math import sqrt, floor
from .io_counter import Counter
counter = Counter()
settings = Settings()


class Node(object):
    """ Nodetype which the R-tree consists of
    """

    def update_MBR(self):
        if len(self.children) == 0 and self.is_root():
            self.mbr = None
            return
        x_min = min([child.mbr.x_min for child in self.children])
        x_max = max([child.mbr.x_max for child in self.children])

        y_min = min([child.mbr.y_min for child in self.children])
        y_max = max([child.mbr.y_max for child in self.children])

        if self.mbr is None:  # Happens first time when root is created, and in splits.
            self.mbr = MBR(x_min, x_max, y_min,y_max)

        else:
            old_mbr = (self.mbr.x_min, self.mbr.x_max, self.mbr.y_min, self.mbr.y_max)
            new_mbr = (x_min, x_max, y_min, y_max)
            if old_mbr != new_mbr:
                self.mbr = MBR(x_min, x_max, y_min, y_max)

                if self.parent is not None:
                    self.parent.update_MBR()
                    counter.write(self.parent)

    def is_leaf(self):
        if len(self.children) == 0:  # Is empty root
            return True
        if all(type(e) is Element for e in self.children):
            return True
        if all(type(e) is Node for e in self.children):
            return False
        raise ValueError("Node is neither Leaf of Internal")

    def is_root(self):
        return self.parent is None

    def get_level(self):
        n = self
        level = 0
        while not n.is_root():
            n = n.parent
            level += 1
        return level

    def get_height(self):
        height = 1
        n = self
        while not n.is_leaf():
            n = n.children[0]
            height += 1
        return height

    def __init__(self, parent=None, children=None):
        if children is None:
            self.children = []
        else:
            self.children = children
        self.mbr = None
        self.parent = parent
        if children is not None:
            self.update_MBR()

    def remove_element(self, E):
        self.children.remove(E)
        self.update_MBR()

    # insert child?
    def insert_element(self, E):
        if len(self.children) < settings.FANOUT:
            E.parent = self
            self.children.append(E)
            self.update_MBR()
            counter.write(self)
        else:  # needs to split node
            L1, L2 = linear_split(self, E)
            if self.is_root():  # if root
                self.children = [L1, L2]
                counter.write(self)
                L1.parent = self
                counter.write(L1.parent)
                L2.parent = self
                counter.write(L2.parent)
                self.update_MBR()
            else:
                try:
                    self.parent.children.remove(self)
                    # TODO, skip this as an update? as it should never fail
                except Exception as e:
                    print(E)
                    raise Exception('Could not remove from parents children list')
                self.parent.insert_element(L1)

                # TODO: Error here?... or it is given that the first one doesn't result in a split. so maybe not..
                self.parent.insert_element(L2)
                # TODO: Missing write? (No it's covered in parent.insert_element)

    def __str__(self):
        # return repr(self)
        return 'Node, children: {}, mbr: {}, has parent:{}'.format(len(self.children), self.mbr, bool(self.parent))


class StarNode(Node):
    """Instance of standard Node, in order to deal with forced reinsertions.
     Also other criterias for split"""

    def is_leaf(self):
        if len(self.children) == 0:  # Is empty root
            return True
        if all(type(e) is Element for e in self.children):
            return True
        if all(type(e) is StarNode for e in self.children):
            return False
        raise ValueError("Node is neither Leaf of Internal")

    def insert_element(self, E, called_levels=None):
        if called_levels is None:
            called_levels = ()

        if len(self.children) < settings.FANOUT:
            E.parent = self
            self.children.append(E)
            self.update_MBR()
            counter.write(self)
        else:  # This is the overflowtreatment
            # print('->', repr(E), self.is_root(), self.is_leaf(), re_inserted)
            if self.get_level() in called_levels:  # SPLIT or self.is_root()?
                # print("star split!")
                L1, L2 = star_split(self, E)

                if self.parent is None:  # if root
                    self.children = [L1, L2]
                    counter.write(self)
                    L1.parent = self
                    counter.write(L1.parent)
                    L2.parent = self
                    counter.write(L2.parent)
                    self.update_MBR()
                if self.parent is not None:
                    try:
                        self.parent.children.remove(self)
                        # TODO, skip this as an update? as it should never fail
                    except Exception as e:
                        print(E)
                        raise Exception('Could not remove from parents children list')
                    self.parent.insert_element(L1)
                    self.parent.insert_element(L2)
                    self.parent.update_MBR()
            else:  # ReInsert
                # print("star reinsert")
                current_level = self.get_level()
                called_levels += (current_level, )

                p = floor(len(self.children) * settings.P_GRADE)  # this should automatically become int..

                p_elements = sorted(self.children + [E],
                                    key=lambda x: -euclidian_distance(x.mbr.center(), self.mbr.center())
                                    )[p:]
                root = self
                while not root.is_root():
                    root = root.parent

                # Delete all first, so no elements are duplicated throuh a split
                for e in p_elements:
                    e.parent = None
                    try:  # causes error if trying to delete new_element from self.
                        self.children.remove(e)
                    except:
                        pass
                counter.write(self)

                if E not in p_elements:  # Else might loose an element
                    self.insert_element(E)

                from .algorithms import insert_star

                for e in p_elements:
                    # TODO: Root can possibly change since last insertion?
                    insert_star(e, root, called_levels=called_levels)


class BF_Node(Node):
    """
    A type of node that handles overflow internally rather than splitting, up to a certain size.
    """
    def is_leaf(self):
        if len(self.children) == 0:  # Is empty root
            return True
        if all(type(e) is Element for e in self.children):
            return True
        if all(type(e) is BF_Node or type(e) is BF_OverflowNode for e in self.children):
            return False
        raise ValueError("Node is neither Leaf of Internal")

    def insert_element(self, E):
        if len(self.children) < settings.FANOUT:
            E.parent = self
            self.children.append(E)
            self.update_MBR()
            counter.write(self)
        elif type(E) is Element:
            # Convert node to an BF_OverflowNode

            if self.is_root():
                self.__class__ = BF_OverflowNode
                self.overflow_blocks = [[] for _ in range(settings.MAX_OVERFLOW)]
                self.overflow_blocks[0] = [] + self.children
                self.insert_element(E)
                self.children = []

            else:
                of_node = BF_OverflowNode(self.children)
                of_node.insert_element(E)
                parent = self.parent
                parent.children.remove(self)
                parent.insert_element(of_node)  # Should never give a split as we just removed an element
        elif type(E) is BF_Node:
            # Doing linear split  (could later be swapped for star..)
            L1, L2 = linear_split(self, E)

            # HACK...
            L1.__class__ = BF_Node
            L2.__class__ = BF_Node
            if self.is_root():  # if root
                self.children = [L1, L2]
                counter.write(self)
                L1.parent = self
                counter.write(L1.parent)
                L2.parent = self
                counter.write(L2.parent)
                self.update_MBR()
            else:
                try:
                    self.parent.children.remove(self)
                    # TODO, skip this as an update? as it should never fail
                except Exception as e:
                    print(E)
                    raise Exception('Could not remove from parents children list')
                self.parent.insert_element(L1)

                # TODO: Error here?... or it is given that the first one doesn't result in a split. so maybe not..
                self.parent.insert_element(L2)
                # TODO: Missing write? (No it's covered in parent.insert_element)
        else:
            raise Exception('Some other case happened')

    def __str__(self):
        # return repr(self)
        return 'BF_Node, children: {}, mbr: {}, has parent:{}'.format(len(self.children), self.mbr, bool(self.parent))


class BF_OverflowNode(BF_Node):

    def __init__(self, first_block):
        super().__init__()

        for e in first_block:
            e.parent = None
        self.overflow_blocks = [[] for _ in range(settings.MAX_OVERFLOW)]
        self.overflow_blocks[0] = first_block

    def update_MBR(self):
        elements = [e for l in self.overflow_blocks for e in l]

        x_min = min([e.mbr.x_min for e in elements])
        x_max = max([e.mbr.x_max for e in elements])

        y_min = min([e.mbr.y_min for e in elements])
        y_max = max([e.mbr.y_max for e in elements])

        if self.mbr is None:  # Happens first time when root is created, and in splits.
            self.mbr = MBR(x_min, x_max, y_min,y_max)

        else:
            old_mbr = (self.mbr.x_min, self.mbr.x_max, self.mbr.y_min, self.mbr.y_max)
            new_mbr = (x_min, x_max, y_min, y_max)
            if old_mbr != new_mbr:
                self.mbr = MBR(x_min, x_max, y_min, y_max)

                if self.parent is not None:
                    self.parent.update_MBR()
                    counter.read(self.parent)  # an read as well?
                    counter.write(self.parent)

    def element_count(self):
        return len([e for l in self.overflow_blocks for e in l])

    def get_height(self):
        return 1

    def is_leaf(self):
        return True

    def insert_element(self, E):
        """
        Inserting into overflow blocks, starting to fill up the first block, and then continuing.
        """
        inserted = False
        for block in self.overflow_blocks:
            counter.read(block)
            if len(block) < settings.FANOUT:
                block.append(E)
                inserted = True
                self.update_MBR()
                counter.write(block)
                break

        if inserted is False:  # If the overflow node is full, a split is needed.
            if self.is_root():

                elements = [e for l in self.overflow_blocks for e in l] + [E]
                tmp_nodes = create_nodes_from_elements(elements)

                bf_nodes = []
                for node in tmp_nodes:
                    bf_node = BF_Node()
                    bf_node.children = node.children + []
                    bf_node.update_MBR()
                    bf_node.parent = self  # set parent after update mbr to not cause unessesary writes
                    bf_nodes.append(bf_node)
                    counter.write(bf_node)

                self.__class__ = BF_Node
                del self.overflow_blocks
                self.children = bf_nodes
                counter.write(self)

                self.update_MBR()

            else:
                # print('Ever here?')
                elements = [e for l in self.overflow_blocks for e in l] + [E]
                tmp_nodes = create_nodes_from_elements(elements)

                root = self
                while not root.is_root():
                    root = root.parent

                self.parent.children.remove(self)
                from .algorithms import insert_bf

                bf_nodes = []
                for node in tmp_nodes:
                    bf_node = BF_Node()
                    bf_node.children = node.children + []
                    bf_node.update_MBR()
                    bf_node.parent = self  # set parent after update mbr to not cause unessesary writes
                    bf_nodes.append(bf_node)
                    counter.write(bf_node)

                for bf_node in bf_nodes:
                    insert_bf(bf_node, root)  # can root change during runtime time?

    def __str__(self):
        # return repr(self)
        return 'OFNode, total_elements: {}, mbr: {}, has parent:{}'.format(len([e for l in self.overflow_blocks for e in l]), self.mbr, bool(self.parent))

class BF_Node_star(Node):
    """
    A type of node that handles overflow internally rather than splitting, up to a certain size.
    """
    def is_leaf(self):
        if len(self.children) == 0:  # Is empty root
            return True
        if all(type(e) is Element for e in self.children):
            return True
        if all(type(e) is BF_Node_star or type(e) is BF_OverflowNode_star for e in self.children):
            return False
        raise ValueError("Node is neither Leaf of Internal")

    def insert_element(self, E):
        if len(self.children) < settings.FANOUT:
            E.parent = self
            self.children.append(E)
            self.update_MBR()
            counter.write(self)
        elif type(E) is Element:
            # Convert node to an BF_OverflowNode

            if self.is_root():
                self.__class__ = BF_OverflowNode_star
                self.overflow_blocks = [[] for _ in range(settings.MAX_OVERFLOW)]
                self.overflow_blocks[0] = [] + self.children
                self.insert_element(E)
                self.children = []

            else:
                of_node = BF_OverflowNode_star(self.children)
                of_node.insert_element(E)
                parent = self.parent
                parent.children.remove(self)
                parent.insert_element(of_node)  # Should never give a split as we just removed an element
        elif type(E) is BF_Node_star:
            # Doing linear split  (could later be swapped for star..)
            L1, L2 = star_split(self, E)

            # HACK...
            L1.__class__ = BF_Node_star
            L2.__class__ = BF_Node_star
            if self.is_root():  # if root
                self.children = [L1, L2]
                counter.write(self)
                L1.parent = self
                counter.write(L1.parent)
                L2.parent = self
                counter.write(L2.parent)
                self.update_MBR()
            else:
                try:
                    self.parent.children.remove(self)
                    # TODO, skip this as an update? as it should never fail
                except Exception as e:
                    print(E)
                    raise Exception('Could not remove from parents children list')
                self.parent.insert_element(L1)

                # TODO: Error here?... or it is given that the first one doesn't result in a split. so maybe not..
                self.parent.insert_element(L2)
                # TODO: Missing write? (No it's covered in parent.insert_element)
        else:
            raise Exception('Some other case happened')

    def __str__(self):
        # return repr(self)
        return 'BF_Node_star, children: {}, mbr: {}, has parent:{}'.format(len(self.children), self.mbr, bool(self.parent))


class BF_OverflowNode_star(BF_Node_star):

    def __init__(self, first_block):
        super().__init__()

        for e in first_block:
            e.parent = None
        self.overflow_blocks = [[] for _ in range(settings.MAX_OVERFLOW)]
        self.overflow_blocks[0] = first_block

    def update_MBR(self):
        elements = [e for l in self.overflow_blocks for e in l]

        x_min = min([e.mbr.x_min for e in elements])
        x_max = max([e.mbr.x_max for e in elements])

        y_min = min([e.mbr.y_min for e in elements])
        y_max = max([e.mbr.y_max for e in elements])

        if self.mbr is None:  # Happens first time when root is created, and in splits.
            self.mbr = MBR(x_min, x_max, y_min,y_max)

        else:
            old_mbr = (self.mbr.x_min, self.mbr.x_max, self.mbr.y_min, self.mbr.y_max)
            new_mbr = (x_min, x_max, y_min, y_max)
            if old_mbr != new_mbr:
                self.mbr = MBR(x_min, x_max, y_min, y_max)

                if self.parent is not None:
                    self.parent.update_MBR()
                    counter.read(self.parent)  # an read as well?
                    counter.write(self.parent)

    def element_count(self):
        return len([e for l in self.overflow_blocks for e in l])

    def get_height(self):
        return 1

    def is_leaf(self):
        return True

    def insert_element(self, E):
        """
        Inserting into overflow blocks, starting to fill up the first block, and then continuing.
        """
        inserted = False
        for block in self.overflow_blocks:
            counter.read(block)
            if len(block) < settings.FANOUT:
                block.append(E)
                inserted = True
                self.update_MBR()
                counter.write(block)
                break

        if inserted is False:  # If the overflow node is full, a split is needed.
            if self.is_root():

                elements = [e for l in self.overflow_blocks for e in l] + [E]
                tmp_nodes = create_nodes_from_elements(elements)

                bf_nodes = []
                for node in tmp_nodes:
                    bf_node = BF_Node_star()
                    bf_node.children = node.children + []
                    bf_node.update_MBR()
                    bf_node.parent = self  # set parent after update mbr to not cause unessesary writes
                    bf_nodes.append(bf_node)
                    counter.write(bf_node)

                self.__class__ = BF_Node_star
                del self.overflow_blocks
                self.children = bf_nodes
                counter.write(self)

                self.update_MBR()

            else:
                elements = [e for l in self.overflow_blocks for e in l] + [E]
                tmp_nodes = create_nodes_from_elements(elements)

                root = self
                while not root.is_root():
                    root = root.parent

                self.parent.children.remove(self)
                from .algorithms import insert_star_bf

                bf_nodes = []
                for node in tmp_nodes:
                    bf_node = BF_Node_star()
                    bf_node.children = node.children + []
                    bf_node.update_MBR()
                    bf_node.parent = self  # set parent after update mbr to not cause unessesary writes
                    bf_nodes.append(bf_node)
                    counter.write(bf_node)

                for bf_node in bf_nodes:
                    insert_star_bf(bf_node, root)  # can root change during runtime time?

    def __str__(self):
        # return repr(self)
        return 'OFNode_star, total_elements: {}, mbr: {}, has parent:{}'.format(len([e for l in self.overflow_blocks for e in l]), self.mbr, bool(self.parent))




class Element(object):
    """Element that are saved in leafnodes of the R-tree"""

    def __init__(self, point, data, parent=None):
        # TODO: Verify data
        # TODO: fix
        self.mbr = MBR(point[0], point[0], point[1], point[1])
        self.data = data
        self.parent = parent

    @property
    def pos(self):
        return (self.mbr.x_max, self.mbr.y_max)

    def __str__(self):
        return "Element, mbr: {}, data: '{}'".format(self.mbr, self.data)


# Helper functions

def create_nodes_from_elements(elements):
    #TODO It counts insertions from here... save state of counter and reseting afterwards?
    current_deque_elements = list(counter.cache)
    current_reads = counter.reads
    current_writes = counter.writes
    from .algorithms import insert
    root = Node()
    for e in elements:
        insert(e, root)

    leaf_nodes = []
    bfs_queue = deque()
    bfs_queue.appendleft(root)
    while bfs_queue:
        current =  bfs_queue.pop()
        if current.is_leaf():
            leaf_nodes.append(current)
        else:
            bfs_queue.extendleft(current.children)

    counter.cache = deque(current_deque_elements, maxlen=settings.CACHE_MAXLEN)
    counter.reads = current_reads
    counter.writes = current_writes
    return leaf_nodes


def create_nodes_from_elements_star(elements):
    #TODO It counts insertions from here... save state of counter and reseting afterwards?
    current_deque_elements = list(counter.cache)
    current_reads = counter.reads
    current_writes = counter.writes
    from .algorithms import insert_star
    root = Node()
    for e in elements:
        insert_star(e, root)

    leaf_nodes = []
    bfs_queue = deque()
    bfs_queue.appendleft(root)
    while bfs_queue:
        current =  bfs_queue.pop()
        if current.is_leaf():
            leaf_nodes.append(current)
        else:
            bfs_queue.extendleft(current.children)

    counter.cache = deque(current_deque_elements, maxlen=settings.CACHE_MAXLEN)
    counter.reads = current_reads
    counter.writes = current_writes
    return leaf_nodes

def euclidian_distance(a, b):
    return sqrt((a[0] - b[0]) ** 2 + (a[1] - b[1]) ** 2)


def linear_split(node, new_element):
    """
    Takes in a iterable of children from the node that are being split, and the new element.
    Returns two nodes, that combined contains all children and new element
    :param children: List of elements that are used to find pair of points to distribute the nodes
    :param new_element: The element originally being inserted
    :return: List of two Nodes
    """
    all_elements = node.children + [new_element]
    pairs = list(combinations(all_elements, 2))
    pairs.sort(key=lambda x: x[0].mbr.distance_to(x[1].mbr))

    most_dist_pair = pairs[-1]

    L1 = Node(children=[most_dist_pair[0]])
    L2 = Node(children=[most_dist_pair[1]])
    most_dist_pair[0].parent = L1
    most_dist_pair[1].parent = L2
    L1.update_MBR()
    L2.update_MBR()

    remaining_elements = [] + all_elements
    remaining_elements.remove(most_dist_pair[0])
    remaining_elements.remove(most_dist_pair[1])

    # TODO: Is there any cases here where there is a split in L1 or L2...
    # Could be possible even tough the first two if sentences.
    # could make them as liste first, and THEN into a node...
    for i, child in enumerate(remaining_elements):
        # Ensuring that one node don't get to few elements...
        if len(L1.children) + (len(remaining_elements) - i) == settings.MIN_FILL:
            L1.insert_element(child)
            continue
        if len(L2.children) + (len(remaining_elements) - i) == settings.MIN_FILL:
            L2.insert_element(child)
            continue

        L1_enlargement = L1.mbr.area_to_circumvent_mbr(child.mbr)
        L2_enlargement = L2.mbr.area_to_circumvent_mbr(child.mbr)

        if L1_enlargement < L2_enlargement:
            L1.insert_element(child)
            continue
        elif L1_enlargement > L2_enlargement:
            L2.insert_element(child)
            continue

        if L1.mbr.area() < L2.mbr.area():
            L1.insert_element(child)
            continue
        elif L1.mbr.area() > L2.mbr.area():
            L2.insert_element(child)
            continue

        if len(L1.children) < len(L2.children):
            L1.insert_element(child)
            continue
        elif len(L1.children) > len(L2.children):
            L2.insert_element(child)
            continue
        else:
            L1.insert_element(child)

    if len(L1.children) + len(L2.children) != settings.FANOUT + 1:
        raise ValueError('An element is missing after a split!')
    return L1, L2


def star_split(node, new_element):
    # The split method from the R-tree algorithm

    all_elements = node.children + [new_element]

    left_sorted_x_dim = sorted(all_elements, key=lambda x: x.mbr.x_min)
    right_sorted_x_dim = sorted(all_elements, key=lambda x: x.mbr.x_max)

    total_x_perimeter = 0
    for pair in generate_distributions(left_sorted_x_dim, settings.FANOUT, settings.MIN_FILL):
        total_x_perimeter += perimeter_value(*pair)
    for pair in generate_distributions(right_sorted_x_dim, settings.FANOUT, settings.MIN_FILL):
        total_x_perimeter += perimeter_value(*pair)

    left_sorted_y_dim = sorted(all_elements, key=lambda x: x.mbr.y_min)
    right_sorted_y_dim = sorted(all_elements, key=lambda x: x.mbr.y_max)

    total_y_perimeter = 0
    for pair in generate_distributions(left_sorted_y_dim, settings.FANOUT, settings.MIN_FILL):
        total_y_perimeter += perimeter_value(*pair)
    for pair in generate_distributions(right_sorted_y_dim, settings.FANOUT, settings.MIN_FILL):
        total_y_perimeter += perimeter_value(*pair)

    split_axis = 'x' if total_x_perimeter < total_y_perimeter else 'y'


    # Now find best split among that axis.. but it doesn't say if i should use right or left split...
    # Thus using both

    if split_axis == 'x':
        distributions = chain(generate_distributions(left_sorted_x_dim, settings.FANOUT, settings.MIN_FILL),
                              generate_distributions(right_sorted_x_dim, settings.FANOUT, settings.MIN_FILL))
    elif split_axis == 'y':
        distributions = chain(generate_distributions(left_sorted_y_dim, settings.FANOUT, settings.MIN_FILL),
                              generate_distributions(right_sorted_y_dim, settings.FANOUT, settings.MIN_FILL))

    best_distribution = ((float('inf'), float('inf')), [])  # ((overlap, area), [dist1, dist2])
    for dist_pair in distributions:
        overlap_val = overlap_value(*dist_pair)
        area_val = area_value(*dist_pair)
        if (overlap_val, area_val) < best_distribution[0]:
            best_distribution = ((overlap_val, area_val), dist_pair)

    L1 = StarNode(children=best_distribution[1][0])
    L1.update_MBR()
    for elem in best_distribution[1][0]:
        elem.parent = L1

    L2 = StarNode(children=best_distribution[1][1])
    L2.update_MBR()
    for elem in best_distribution[1][1]:
        elem.parent = L2

    return L1, L2


def generate_distributions(sorted_list, max_fill, min_fill):
    # Returns a list with pairs of lists, where each pair is a split of the distribution

    num_distributions = max_fill - (2 * min_fill) + 2  # todo: check if could be +3...
    for k in range(num_distributions):
        yield (sorted_list[:min_fill+k], sorted_list[min_fill+k:])


def perimeter_value(first_group, second_group):  # equals margin-value in paper
    return MBR.create_circumventing_mbr([elem.mbr for elem in first_group]).perimeter() + \
        MBR.create_circumventing_mbr([elem.mbr for elem in second_group]).perimeter()


def area_value(first_group, second_group):
    first_group_mbr = MBR.create_circumventing_mbr([elem.mbr for elem in first_group])
    second_group_mbr = MBR.create_circumventing_mbr([elem.mbr for elem in second_group])

    return first_group_mbr.area() + second_group_mbr.area()


def overlap_value(first_group, second_group):
    first_group_mbr = MBR.create_circumventing_mbr([elem.mbr for elem in first_group])
    second_group_mbr = MBR.create_circumventing_mbr([elem.mbr for elem in second_group])

    return first_group_mbr.overlap(second_group_mbr)
