import pytest
from .mbr import MBR
from math import isclose


def test_calculate_area_of_1():
    mbr = MBR(0, 1, 0, 1)
    assert mbr.area() == 1


def test_negative_value_gives_positive_area():
    mbr = MBR(-1, 0, 0, 1)
    assert mbr.area() == 1


def test_initialize_with_min_larger_than_max_raises_error():
    with pytest.raises(ValueError):
        MBR(-1, -2, 0, 0)


def test_contains_point_returns_true_for_point_that_are_within():
    mbr = MBR(0, 3, 0, 3)
    assert mbr.contains_point((1, 1))


def test_contains_point_returns_true_for_point_that_are_on_border():
    mbr = MBR(0, 3, 0, 3)
    assert mbr.contains_point((3, 3))


def test_area_to_circumvent_point_1_2_from_1x1_is_1():
    mbr = MBR(0, 1, 0, 1)
    assert mbr.area_to_circumvent_point((1, 2)) == 1


def test_area_to_circumvent_point_3_3_from_1x1_is_8():
    mbr = MBR(0, 1, 0, 1)
    assert mbr.area_to_circumvent_point((3, 3)) == 8


def test_area_to_circumvent_point_inside_is_0():
    mbr = MBR(0, 2, 0, 2)
    assert mbr.area_to_circumvent_point((1, 1)) == 0


def test_area_to_circumvent_mbr_adjacent_1x1():
    mbr_a = MBR(0, 1, 0, 1)
    mbr_b = MBR(1, 2, 0, 1)
    assert mbr_a.area_to_circumvent_mbr(mbr_b) == 1
    assert mbr_b.area_to_circumvent_mbr(mbr_a) == 1


def test_area_to_circumvent_mbr_with_other_shape():
    mbr_a = MBR(0, 1, 0, 1)
    mbr_b = MBR(1, 2, 0, 2)
    assert mbr_a.area_to_circumvent_mbr(mbr_b) == 3
    assert mbr_b.area_to_circumvent_mbr(mbr_a) == 2


def test_area_to_circumvent_mbr_that_cirvumvents():
    mbr_inner = MBR(1, 2, 1, 2)
    mbr_outer = MBR(0, 3, 0, 3)
    assert mbr_inner.area_to_circumvent_mbr(mbr_outer) == 8
    assert mbr_outer.area_to_circumvent_mbr(mbr_inner) == 0


def test_overlaps_overlaps_itself():
    mbr_a = MBR(0, 2, 0, 2)
    assert mbr_a.overlaps(mbr_a)


def test_overlaps_returns_true_when_corner_overlaps():
    mbr_a = MBR(0, 2, 0, 2)
    mbr_b = MBR(1, 3, 1, 3)
    assert mbr_a.overlaps(mbr_b) == True
    assert mbr_b.overlaps(mbr_a) == True


def test_overlaps_returns_true_when_sharing_border():
    # vertical border
    mbr_a = MBR(0, 2, 0, 2)
    mbr_b = MBR(2, 4, 0, 2)
    assert mbr_a.overlaps(mbr_b) == True
    assert mbr_b.overlaps(mbr_a) == True

    # horontisonal border
    mbr_a = MBR(0, 2, 0, 2)
    mbr_b = MBR(0, 2, 2, 4)
    assert mbr_a.overlaps(mbr_b) == True
    assert mbr_b.overlaps(mbr_a) == True


def test_overlaps_returns_true_when_only_sharing_corner():
    # is true when they only share corner, is diagonal
    #  __| |
    # | |
    mbr_a = MBR(0, 2, 0, 2)
    mbr_b = MBR(2, 4, 2, 4)
    assert mbr_a.overlaps(mbr_b) == True
    assert mbr_b.overlaps(mbr_a) == True


def test_overlaps_returns_true_when_one_cirumvents_the_other_completly():
    mbr_a = MBR(0, 4, 0, 4)
    mbr_b = MBR(1, 3, 1, 3)
    assert mbr_a.overlaps(mbr_b) == True
    assert mbr_b.overlaps(mbr_a) == True


def test_overlaps_returns_false_when_no_overlap():
    mbr_a = MBR(0, 4, 0, 4)
    mbr_b = MBR(5, 9, 7, 9)
    assert mbr_a.overlaps(mbr_b) == False
    assert mbr_b.overlaps(mbr_a) == False


def test_point_overlaps_point():
    mbr_a = MBR(0, 0, 1, 1)
    mbr_b = MBR(0, 0, 1, 1)

    assert mbr_a.overlaps(mbr_b)
    assert mbr_b.overlaps(mbr_a)


def test_corners_is_4_points_when_mbr_is_square():
    mbr = MBR(0, 1, 0, 1)
    assert len(mbr.corners()) == 4


def test_corners_is_2_points_when_mbr_is_line():
    mbr = MBR(0, 0, 0, 1)
    assert len(mbr.corners()) == 2


def test_corners_is_1_points_when_mbr_is_point():
    mbr = MBR(0, 0, 0, 0)
    assert len(mbr.corners()) == 1


def test_distance_to_overlapping_mbrs():
    mbr_a = MBR(0, 4, 0, 4)
    mbr_b = MBR(1, 3, 1, 3)
    assert mbr_a.distance_to(mbr_b) == 0
    assert mbr_b.distance_to(mbr_a) == 0


def test_distance_to_diagonal_mbrs():
    # Here the shortest pair of cornerpoints between the mbr counts.
    mbr_a = MBR(0, 1, 0, 1)
    mbr_b = MBR(2, 3, 2, 3)
    assert isclose(mbr_a.distance_to(mbr_b), 1.4142, rel_tol=0.0001)
    assert isclose(mbr_b.distance_to(mbr_a), 1.4142, rel_tol=0.0001)


def test_distance_to_adjacent_nonoverlapping_mbrs():
    mbr_a = MBR(0, 1, 0, 4)
    mbr_b = MBR(4, 5, 2, 3)

    assert mbr_a.distance_to(mbr_b) == 3
    assert mbr_b.distance_to(mbr_a) == 3


def test_overlap_between_mbrs_where_one_circumvents():
    mbr_a = MBR(0, 4, 0, 4)
    mbr_b = MBR(1, 3, 1, 3)
    assert mbr_a.overlap(mbr_b) == 4
    assert mbr_b.overlap(mbr_a) == 4


def test_overlap_between_diagonal_mbrs_that_dont_overlap():
    mbr_a = MBR(0, 1, 0, 4)
    mbr_b = MBR(1, 3, 4, 9)

    assert mbr_a.overlap(mbr_b) == 0
    assert mbr_b.overlap(mbr_a) == 0


def test_overlap_between_mbrs_partly_overlap_side():
    mbr_a = MBR(0, 4, 0, 4)
    mbr_b = MBR(3, 5, 1, 3)

    assert mbr_a.overlap(mbr_b) == 2
    assert mbr_b.overlap(mbr_a) == 2


def test_overlap_between_mbrs_partly_overlap_corner():
    mbr_a = MBR(0, 4, 0, 4)
    mbr_b = MBR(2, 5, 2, 5)

    assert mbr_a.overlap(mbr_b) == 4
    assert mbr_b.overlap(mbr_a) == 4


def test_static_get_mbr():
    x_values = [1, 2, 3, -19, 20, 4]
    y_values = [2, 4, -1, 6, -39, 300]

    mbr = MBR.create_mbr(x_values=x_values, y_values=y_values)
    assert type(mbr) is MBR
    assert mbr.x_min == min(x_values)
    assert mbr.x_max == max(x_values)
    assert mbr.y_min == min(y_values)
    assert mbr.y_max == max(y_values)


def test_static_create_circumventing_mbr():
    mbr_a = MBR(0, 4, 0, 4)
    mbr_b = MBR(2, 5, 2, 5)
    mbr_c = MBR(2, 5, -1, 5)
    mbr = MBR.create_circumventing_mbr([mbr_a, mbr_b, mbr_c])

    assert mbr.x_min == 0
    assert mbr.x_max == 5
    assert mbr.y_min == -1
    assert mbr.y_max == 5


def test_mbr_perimeter_positive_values():
    mbr = MBR(1, 4, 1, 4)

    assert mbr.perimeter() == 12


def test_mbr_peremeter_nevative_values():
    mbr = MBR(-4, -1, 1, 4)

    assert mbr.perimeter() == 12


def test_mbr_center_positive_values():
    mbr = MBR(0, 4, 0, 4)

    assert mbr.center() == (2, 2)


def test_mbr_center_negative_values():
    mbr = MBR(-4, -2, -4, -2)

    assert mbr.center() == (-3, -3)


def test_mbr_center_for_point():
    mbr = MBR(1, 1, 1, 1)

    assert mbr.center() == (1, 1)


def test_center_for_origo():
    mbr = MBR(0, 0, 0, 0)

    assert mbr.center() == (0, 0)
